#!/usr/bin/env python3

# standards
from datetime import timedelta
from pathlib import Path

# 3rd parties
import pytest
from requests import Response

# hublot
from hublot import Client
from hublot.cache.storage import DiskStorage
from hublot.logs import LogEntry
from .utils import assert_responses_equal, dummy_prepared_request, dummy_response


def iter_pairs(client):
    for url in ('http://one/', 'http://two/'):
        for method, body in [('GET', None), ('POST', b'dummy body')]:
            for headers in ({}, {'X-Test': '1'}, {'X-Test': '2'}):
                # NB there's more comprehensive tests for cache key equivalent in `test_cache_keys.py`, but here's a sample
                request_kwargs = {'url': url, 'method': method, 'headers': headers, 'data': body}
                preq = dummy_prepared_request(client, **request_kwargs)
                res = dummy_response(preq)
                yield request_kwargs, res


def test_cache(reinstantiable_client):
    client = reinstantiable_client()
    pairs = [
        (dummy_prepared_request(client, **req), res)
        for req, res in iter_pairs(client)
    ]
    log_entries = [
        LogEntry(preq)
        for preq, _res_unused in pairs
    ]
    cache = client.cache
    for (preq, _res_unused), log in zip(pairs, log_entries):
        assert cache.get(preq, log) is None  # else test is invalid
        assert log.cached is False
    for (preq, res), log in zip(pairs, log_entries):
        cache.put(preq, log, res)
    cache = reinstantiable_client().cache
    for (preq, res), log in zip(pairs, log_entries):
        assert_responses_equal(cache.get(preq, log), res)
        assert log.cached is True


def test_client_caching(mocker, reinstantiable_client):
    client = reinstantiable_client()
    for req, res in iter_pairs(client):
        request = mocker.patch.object(client.session, 'request', return_value=res)
        assert_responses_equal(client.fetch(**req), res)
        request.assert_called_once()
    client = reinstantiable_client()
    for req, res in iter_pairs(client):
        request = mocker.patch.object(client.session, 'request', return_value=res)
        assert_responses_equal(client.fetch(**req), res)
        request.assert_not_called()


def test_http_errors_are_cached(client, server):
    one = client.get(f'{server}/fail-with-random-value', raise_for_status=False)
    assert one.status_code == 500
    two = client.get(f'{server}/fail-with-random-value', raise_for_status=False)
    assert two.status_code == 500
    three = client.get(f'{server}/fail-with-random-value', raise_for_status=False, force_cache_stale=True)
    assert three.status_code == 500
    assert one.text == two.text
    assert two.text != three.text


def test_repeated_http_headers_are_cached(reinstantiable_client, server):
    client = reinstantiable_client()
    res = client.get(f'{server}/cookies/set-two-cookies')
    assert res.headers.get('Set-Cookie') == 'a=1, b=2'
    assert res.raw._fp.headers.get_all('Set-Cookie') == ['a=1', 'b=2']   # pylint: disable=protected-access
    unique = res.text

    client = reinstantiable_client()
    res = client.get(f'{server}/cookies/set-two-cookies')
    assert res.text == unique  # check that it was cached
    assert res.headers.get('Set-Cookie') == 'a=1, b=2'
    #if callable(getattr(res.headers, 'get_all', None)):
    assert res.headers.get_all('Set-Cookie') == ['a=1', 'b=2']

    assert [f'{c.name}={c.value!r}' for c in client.session.cookies] == ["a='1'", "b='2'"]


def test_cache_wont_save_body_with_wrong_length(client):
    preq = dummy_prepared_request(
        client,
        data=b'More than one byte',
    )
    preq.headers['Content-Length'] = '1'
    response = Response()
    response.request = preq
    with pytest.raises(Exception) as ex:
        client.cache.put(preq, LogEntry(preq), response)
    assert 'body has 18 bytes but Content-Length is 1' in str(ex)


def test_cache_wont_save_get_request_with_content_length(client):
    preq = dummy_prepared_request(
        client,
        method='GET',
    )
    preq.headers['Content-Length'] = '999'
    response = Response()
    response.request = preq
    with pytest.raises(Exception) as ex:
        client.cache.put(preq, LogEntry(preq), response)
    assert 'body has 0 bytes but Content-Length is 999' in str(ex)


def test_cache_can_handle_empty_post_request(client, server):
    for _from_cache_unused in (False, True):
        res = client.fetch(f'{server}/echo', data={})
        obtained = res.json()
        obtained.pop('headers')
        assert obtained == {'method': 'POST', 'args': {}, 'files': {}, 'form': {}, 'json': None}


def test_cant_pass_cache_kwargs_and_preinstantiated_cache(cache):
    with pytest.raises(Exception) as ex:
        Client(cache=cache, max_cache_age=timedelta(10))
    assert "can't specify a max_age" in str(ex)


def test_cache_can_be_specified_as_path():
    client = Client(cache=Path('/cache'))
    assert isinstance(client.cache.storage, DiskStorage)
    assert client.cache.storage.root_path == Path('/cache')


def test_cache_cannot_be_specified_as_str():
    # The idea of not accepting strings is mostly to just keep the client code readable, and maybe preserve forward flexibility.
    # But if this ever a problem, we could consider auto-converting str objects to Path objects.
    with pytest.raises(ValueError):
        Client(cache='/cache')


def test_from_cache_attribute(client, server):
    for from_cache in (False, True):
        res = client.fetch(f'{server}/hello')
        assert res.from_cache == from_cache


def test_cache_can_handle_empty_http_reason(client, server):
    for _from_cache_unused in (False, True):
        res = client.fetch(f'{server}/no-reason')
        assert res.reason == ''
        assert res.text == 'hello'



#!/usr/bin/env python3

# standards
from io import BytesIO
from itertools import combinations, product
from typing import Dict, Optional

# 3rd parties
from requests import PreparedRequest, Request, Response
from requests.structures import CaseInsensitiveDict

# hublot
from hublot import Client


def dummy_prepared_request(client: Client, **kwargs):
    url = kwargs.pop('url', 'http://example.com/test')
    if not isinstance(url, Request):
        kwargs.setdefault('method', 'POST')
        if kwargs['method'] in ('POST', 'PUT'):
            kwargs.setdefault('data', b'This is my request data')
    return client._prepare(client.build_request(url, **kwargs))  # pylint: disable=protected-access


def dummy_response(
    preq: PreparedRequest,
    status_code: int = 200,
    reason: str = 'OK',
    headers: Optional[Dict[str, str]] = None,
    data: bytes = b'This is my response data',
):
    res = Response()
    res.request = preq
    res.status_code = status_code
    res.reason = reason
    res.headers = CaseInsensitiveDict(headers or {})
    res.url = preq.url  # type: ignore
    res.raw = BytesIO(data)
    return res


def iter_nonequal_pairs(equivalencies):
    for index, group in enumerate(equivalencies):
        all_other_elems = [
            other_group[0]
            for other_index, other_group in enumerate(equivalencies)
            if other_index > index  # Using > to avoid wastefully comparing A to B and B to A
        ]
        for elem, other_elem in product(group, all_other_elems):
            yield elem, other_elem


def iter_equal_pairs(equivalencies):
    for group in equivalencies:
        if len(group) > 2:
            for elem_1, elem_2 in combinations(group, 2):
                yield elem_1, elem_2


def assert_responses_equal(res1, res2):
    state1 = res1.__getstate__()
    state2 = res2.__getstate__()
    state1['request'] = state1['request'] and state1['request'].__dict__
    state2['request'] = state2['request'] and state2['request'].__dict__
    try:
        assert state1 == state2
    except AssertionError:  # pragma: no cover, if the tests pass then this won't get called
        print(state1)
        print(state2)
        raise



#!/usr/bin/env python3

# standards
from datetime import datetime, timedelta
import gzip
from pathlib import Path
import re
from typing import Iterable, Optional

# 3rd parties
from requests import Response

# hublot
from .binaryblob import compose_binary_blob, parse_binary_blob
from .key import CacheKey


class Storage:  # pragma: no cover

    def read(self, key: CacheKey, max_age: Optional[timedelta] = None) -> Optional[Response]:
        raise NotImplementedError

    def write(self, key: CacheKey, response: Response) -> None:
        raise NotImplementedError

    def iter_all_keys(self) -> Iterable[CacheKey]:
        raise NotImplementedError

    def prune(self, max_age: timedelta):
        raise NotImplementedError


class DiskStorage(Storage):

    def __init__(self, root_path: Path):
        self.root_path = root_path

    def read(self, key: CacheKey, max_age: Optional[timedelta] = None) -> Optional[Response]:
        file_path = self._file_path(key)
        if not file_path.exists():
            return None
        if max_age is not None:
            file_age = current_datetime() - datetime.fromtimestamp(file_path.stat().st_mtime)
            if file_age > max_age:
                return None
        with gzip.open(file_path, 'rb') as file_in:
            return parse_binary_blob(file_in.read())

    def write(self, key: CacheKey, response: Response) -> None:
        file_path = self._file_path(key)
        file_path.parent.mkdir(parents=True, exist_ok=True)
        with gzip.open(file_path, 'wb') as file_out:
            file_out.write(compose_binary_blob(response))

    def iter_all_keys(self) -> Iterable[CacheKey]:
        for file_path in self._iter_all_files():
            parts = list(file_path.relative_to(self.root_path).parts)
            parts[-1] = re.sub(r'\.gz$', '', file_path.name)
            yield CacheKey.from_path_parts(parts)

    def prune(self, max_age: timedelta):
        now = current_datetime()
        dirs_to_check = set()
        for file_path in self._iter_all_files():
            file_age = now - datetime.fromtimestamp(file_path.stat().st_mtime)
            if file_age > max_age:
                file_path.unlink()
                dirs_to_check.add(file_path.parent)
        for dir_path in dirs_to_check:
            while not next(dir_path.iterdir(), None):
                dir_path.rmdir()
                dir_path = dir_path.parent

    def _iter_all_files(self) -> Iterable[Path]:
        return self.root_path.glob('**/*.gz')

    def _file_path(self, key: CacheKey) -> Path:
        file_path = self.root_path / Path(*key.path_parts)
        if not file_path.suffix == '.gz':
            file_path = file_path.parent / f'{file_path.name}.gz'
        return file_path


def current_datetime():
    # This is put in a separate function so that tests can patch that function
    return datetime.now()



#!/usr/bin/env python3

# 3rd parties
from requests import Request

# hublot
from hublot import RequestableABC


def test_requestable_is_runtime_type():
    """ `hublot.RequestableABC` is not a type annotation, it's an ABC class that can be used with `isinstance` """
    assert isinstance('http://blah/', RequestableABC)
    assert isinstance(Request('GET', 'http://blah/'), RequestableABC)



#!/usr/bin/env python3

# standards
from contextlib import contextmanager
from datetime import timedelta
from time import sleep, time
from typing import Dict, Optional
from urllib.parse import urlparse

# 3rd parties
from requests import PreparedRequest

# hublot
from .logs import LogEntry


class CourtesySleep:

    def __init__(self, courtesy_sleep: Optional[timedelta]):
        if courtesy_sleep is None:
            courtesy_sleep = timedelta(0)
        self.courtesy_sleep = courtesy_sleep
        self.last_request_per_host: Dict[str, float] = {}

    @contextmanager
    def __call__(self, preq: PreparedRequest, log: LogEntry, courtesy_sleep: Optional[timedelta] = None):
        if courtesy_sleep is None:
            courtesy_sleep = self.courtesy_sleep
        host = str(urlparse(preq.url).hostname)
        last_request = self.last_request_per_host.get(host, 0)
        delay_seconds = (last_request + courtesy_sleep.total_seconds()) - time()
        if delay_seconds > 0:
            log.courtesy_seconds = delay_seconds
            sleep(delay_seconds)
        try:
            yield
        finally:
            # NB we store the time after the request is complete
            self.last_request_per_host[host] = time()



#!/usr/bin/env python3

# hublot
from hublot import Client


def test_cookies_are_saved(client, server):
    assert client.get(f'{server}/cookies/get').json() == {}
    client.get(f'{server}/cookies/set?coo=kie')
    assert client.get(f'{server}/cookies/get').json() == {'coo': 'kie'}


def test_cookies_are_available_via_client(cache, server):
    for _ in (1, 2):
        client = Client(cache=cache)  # keep cache, clear cookies
        client.get(f'{server}/cookies/set?coo=kie')
        assert dict(client.cookies) == {'coo': 'kie'}


def test_cookies_are_available_via_session(cache, server):
    for _ in (1, 2):
        client = Client(cache=cache)  # keep cache, clear cookies
        client.get(f'{server}/cookies/set?coo=kie')
        assert dict(client.session.cookies) == {'coo': 'kie'}


def test_cookies_are_available_via_response(cache, server):
    for _ in (1, 2):
        client = Client(cache=cache)  # keep cache, clear cookies
        response = client.get(f'{server}/cookies/set?coo=kie')
        assert dict(response.cookies) == {'coo': 'kie'}


def test_cookies_are_set_when_running_from_cache(reinstantiable_client, server):
    for _ in (1, 2):
        client = reinstantiable_client()
        assert client.get(f'{server}/cookies/get').json() == {}
        client.get(f'{server}/cookies/set?coo=kie')
        assert client.get(f'{server}/cookies/get').json() == {'coo': 'kie'}


def test_cached_redirects(cache, server):
    for _ in (1, 2):
        client = Client(cache=cache)
        client.get(f'{server}/redirect/chain/1')
        cookies = {c.name: c.value for c in client.session.cookies}
        assert cookies == {'redirect1': 'yes', 'redirect2': 'yes', 'redirect3': 'yes'}


def test_cookies_can_be_disabled(cache, server):
    for _ in (1, 2):
        client = Client(cache=cache, cookies_enabled=False)
        assert client.get(f'{server}/cookies/get').json() == {}
        client.get(f'{server}/cookies/set?coo=kie')
        assert client.get(f'{server}/cookies/get').json() == {}


def test_multiple_cookies_headers(reinstantiable_client, server):
    for reading_from_cache in (False, True):
        client = reinstantiable_client()
        response = client.get(f'{server}/cookies/set-two-cookies')
        if not reading_from_cache:
            # Check that the mock server returns cookies in 2 separate headers, else this test isn't testing what it meant to test.
            # Can't run this test on the second iteration, because we don't have `response.raw` when coming from cache.
            response_cookies = sorted(
                (key, value)
                for key, value in response.raw.headers.items()
                if key == 'Set-Cookie'
            )
            assert response_cookies == [('Set-Cookie', 'a=1'), ('Set-Cookie', 'b=2')]
        assert sorted(response.cookies.items()) == [('a', '1'), ('b', '2')]
        assert sorted(client.session.cookies.items()) == [('a', '1'), ('b', '2')]


def test_response_that_redirects_to_same_url_with_cookie(client, server):
    assert client.get(f'{server}/self-redirect').text == 'ok'


def test_set_cookie_manually(client, server):
    client.cookies.set('x', '1')
    assert client.get(f'{server}/cookies/get').json() == {'x': '1'}



#!/usr/bin/env python3

# standards
from abc import ABC
from contextlib import contextmanager
from datetime import timedelta
from pathlib import Path
import re
from typing import Any, Dict, Optional, Set, Union
from urllib.parse import urljoin

# 3rd parties
from requests import PreparedRequest, Request, Response, Session, TooManyRedirects
from requests.cookies import MockRequest, RequestsCookieJar
import urllib3.util.url

# hublot
from .cache import Cache, CacheKey, UserSpecifiedCacheKey
from .courtesy import CourtesySleep
from .decorator import SCRAPER_LOCAL
from .logs import LOGGER, LogEntry
from .utils import HublotCookiePolicy, MockResponse
from .version import HUBLOT_VERSION


MAX_REDIRECTS = 10


class RequestableABC(ABC):
    """ Umbrella type for values that can be passed to the `request` method """

RequestableABC.register(str)
RequestableABC.register(Request)

# Sadly mypy doesn't understand ABC.register, so AFAICT we need `RequestableABC` for run-time `isinstance` checks, and
# `Requestable` for static analysis type checks
Requestable = Union[RequestableABC, str, Request]


class Client:

    def __init__(
        self,
        cache: Optional[Union[Cache, Path, str]] = None,
        courtesy_sleep: Optional[Union[CourtesySleep, timedelta]] = timedelta(seconds=5),
        session: Optional[Session] = None,
        max_cache_age: Optional[timedelta] = None,
        user_agent: str = f'hublot/{HUBLOT_VERSION}',
        cookies_enabled: bool = True,
        proxies: Optional[Dict[str, str]] = None,
    ):
        self.cache = Cache.load(cache, max_cache_age)
        if not isinstance(courtesy_sleep, CourtesySleep):
            courtesy_sleep = CourtesySleep(courtesy_sleep)  # malkovitch malkovitch
        self.courtesy_sleep = courtesy_sleep
        self.session = session or Session()
        self.session.cookies.set_policy(HublotCookiePolicy(cookies_enabled))
        if proxies:
            self.session.proxies = proxies
        self.user_agent = user_agent

    @property
    def cookies(self) -> RequestsCookieJar:
        return self.session.cookies

    def fetch(  # pylint: disable=too-many-locals
        self,
        url: Requestable,
        courtesy_sleep: Optional[timedelta] = None,
        raise_for_status: bool = True,
        force_cache_stale: bool = False,
        allow_redirects: bool = True,
        cache_key: Optional[UserSpecifiedCacheKey] = None,
        max_cache_age: Optional[timedelta] = None,
        proxies: Optional[Dict[str, str]] = None,
        verify: Optional[bool] = True,
        _redirected_from: Optional[Response] = None,
        **kwargs,
    ) -> Response:
        """
        NB the first parameter is called `url`, even though it could also be a `Request` object, simply because that's shorter and
        nicer than `requestable`. Hey, `urllib.request.urlopen` does the same thing, so.
        """
        frame = SCRAPER_LOCAL.stack[-1]
        if frame.is_retry:
            force_cache_stale = True
            courtesy_sleep = timedelta(0)
        with patched_encode_invalid_chars():
            req = self.build_request(url, **kwargs)
            preq = self._prepare(req)
            log = LogEntry(preq, is_redirect=(_redirected_from is not None))
            res = self._fetch_response(
                req,
                preq,
                log,
                courtesy_sleep,
                force_cache_stale,
                cache_key,
                max_cache_age,
                proxies,
                verify,
            )
            if res.from_cache:  # type: ignore[attr-defined]  # we add that attribute
                self.session.cookies.extract_cookies(MockResponse(res), MockRequest(preq))  # type: ignore[arg-type]
        if _redirected_from:
            res.history = [*_redirected_from.history, _redirected_from]
        LOGGER.info('%s', log)
        return self._handle_response(
            preq,
            res,
            raise_for_status,
            force_cache_stale,
            allow_redirects,
            cache_key,
            max_cache_age,
            _redirected_from,
        )

    @staticmethod
    def build_request(url: Requestable, **kwargs) -> Request:
        """
        Takes a `Requestable` and compiles it into a `requests.Request` object.
        """
        if isinstance(url, Request):
            if kwargs:
                # I'm putting this here to avoid unforeseen behaviour, but if it ever becomes a problem, could consider relaxing
                # this constraint.
                kwargs_str = ', '.join(sorted(map(repr, kwargs)))
                raise TypeError(f"If passing a Request object to `fetch`, can't specify kwargs: {kwargs_str}")
            return url
        else:
            method = kwargs.pop(
                'method',
                'POST' if (
                    kwargs.get('data') is not None
                    or kwargs.get('files') is not None
                    or kwargs.get('json') is not None
                )
                else 'GET'
            ).upper()
            return Request(method, url, **kwargs)

    def _prepare(self, req: Request) -> PreparedRequest:
        """
        Given the user-supplied arguments to the `request`, method, compile a `PreparedRequest` object. Normally this is done
        within Requests (and it will still be done by Requests when we call it), but we need this in order to compute the cache
        key.

        NB this modifies the given `params` dict to remove the 'method', if any.
        """
        req.data = self._read_files_to_memory(req.data)
        req.files = self._read_files_to_memory(req.files)
        req.headers.setdefault('User-Agent', self.user_agent)
        return self.session.prepare_request(req)

    def _fetch_response(
        self,
        req: Request,
        preq: PreparedRequest,
        log: LogEntry,
        courtesy_sleep: Optional[timedelta],
        force_cache_stale: bool,
        cache_key: Optional[UserSpecifiedCacheKey],
        max_cache_age: Optional[timedelta],
        proxies: Optional[Dict[str, str]],
        verify: Optional[bool],
    ) -> Response:
        """
        Either read the Response from cache, or perform the HTTP transaction and save the response to cache
        """
        if self.cache and not force_cache_stale:
            res = self.cache.get(preq, log, cache_key, max_cache_age)
            if res is not None:
                res.from_cache = True  # type: ignore[attr-defined]  # we add that attribute
                return res
        with self.courtesy_sleep(preq, log, courtesy_sleep):
            res = self.session.request(
                allow_redirects=False,
                proxies=proxies,
                verify=verify,
                **req.__dict__,
            )
        res.from_cache = False  # type: ignore[attr-defined]  # we add that attribute
        if self.cache:
            self.cache.put(preq, log, res, cache_key)
        return res

    def _handle_response(
        self,
        preq: PreparedRequest,
        res: Response,
        raise_for_status: bool,
        force_cache_stale: bool,
        allow_redirects: bool,
        cache_key: Optional[UserSpecifiedCacheKey],
        max_cache_age: Optional[timedelta],
        _redirected_from: Optional[Response],
    ):
        """
        Examine the `Response` we just got, and decide what to do with it: follow a redirect (if any), raise an exception, or just
        return the response.
        """
        if allow_redirects and res.is_redirect:
            if _redirected_from and len(_redirected_from.history) >= MAX_REDIRECTS:
                raise TooManyRedirects(f'Exceeded {MAX_REDIRECTS} redirects')
            return self.fetch(
                urljoin(preq.url, res.headers['Location']),  # type: ignore  # we've checked that `preq.url` is not None
                courtesy_sleep=timedelta(0),
                raise_for_status=raise_for_status,
                force_cache_stale=force_cache_stale,
                cache_key=cache_key and CacheKey.parse(cache_key).next_in_sequence(),
                max_cache_age=max_cache_age,
                _redirected_from=res,
            )
        if raise_for_status:
            res.raise_for_status()
        return res

    def _read_files_to_memory(self, obj: Any) -> Any:
        """
        If `data` or `files` are open file objects, read their contents to memory. We need to see their data in order to include the
        request body in the cache key.
        """
        if callable(getattr(obj, 'read', None)):
            body = obj.read()
            obj.close()
            return body
        elif isinstance(obj, dict):
            return {key: self._read_files_to_memory(value) for key, value in obj.items()}
        else:
            return obj

    ### for a thin layer of Requests compatibility

    def request(self, method: str, url: str, **kwargs) -> Response:
        return self.fetch(url, method=method, **kwargs)

    def get(self, url: str, **kwargs) -> Response:
        return self.fetch(url, method='GET', **kwargs)

    def post(self, url: str, data=None, **kwargs) -> Response:
        return self.fetch(url, method='POST', data=data, **kwargs)


_encode_invalid_chars = urllib3.util.url._encode_invalid_chars  # pylint: disable=protected-access


def _encode_invalid_chars_preserve_case(component: str, allowed_chars: Set[str], encoding: str = 'utf-8') -> str:
    new = _encode_invalid_chars(component, allowed_chars, encoding=encoding)
    if new != component:
        # `urllib3`, which is used internally by `requests`, changes the casing of %xx escapes, always forcing them to uppercase.
        # This means that if a server is set up such that when you ask for a URL with uppercase escapes it redirects you to the
        # same URL with lowercase escapes, then `requests` is unable to fetch that page -- the URL in the redirect response
        # 'location' will be re-uppercased, leading to a redirect loop. Here we work around that by chaging it back to lowercase.
        #
        # You could argue that the HTTP spec says %-escapes are case-insensitive, so it's the server described above that's
        # buggy, but browsers have no problem fetching such a page, so I'd argue it's a `requests` bug.
        #
        new = re.sub(r'%[0-9a-fA-F]{2}', lambda m: m.group().lower(), new)
    return new


@contextmanager
def patched_encode_invalid_chars():
    # pylint: disable=protected-access
    original = urllib3.util.url._encode_invalid_chars
    urllib3.util.url._encode_invalid_chars = _encode_invalid_chars_preserve_case
    try:
        yield
    finally:
        urllib3.util.url._encode_invalid_chars = original



#!/usr/bin/env python3

from .cache import Cache, CacheKey
from .client import Client, Requestable, RequestableABC
from .courtesy import CourtesySleep
from .decorator import retry_on_scraper_error
from .exceptions import ScraperError
from .logs import LOGGER as logger
from .logs import basic_logging_config
from .utils import lookup



#!/usr/bin/env python3

# standards
from http.cookiejar import CookiePolicy
from typing import Callable, Dict, TypeVar

# 3rd parties
from requests import Response

# hublot
from .exceptions import ScraperError


class HublotCookiePolicy(CookiePolicy):  # pragma: no cover, we don't care how this gets called, as long as it works

    netscape = True
    rfc2965 = False
    hide_cookie2 = False

    def __init__(self, cookies_enabled: bool):
        super().__init__()
        self.cookies_enabled = cookies_enabled

    def set_ok(self, cookie, request):
        return self.cookies_enabled

    def return_ok(self, cookie, request):
        return self.cookies_enabled

    def domain_return_ok(self, domain, request):
        return self.cookies_enabled

    def path_return_ok(self, path, request):
        return self.cookies_enabled


class MockResponse: # pragma: no cover -- again, as long as it works, we don't care how urllib3 or whatever it is calls this

    def __init__(self, response: Response):
        self.response = response

    def info(self):
        return self

    def get_all(self, name, failobj=None):
        # This is a mock of `email.message.EmailMessage.get_all` -- ``Return a list of all the values for the field named name. If
        # there are no such named headers in the message, failobj is returned''
        #
        # See https://docs.python.org/3.8/library/email.message.html#email.message.EmailMessage.get_all
        if callable(getattr(self.response.headers, 'get_all', None)):
            # It must be that `headers` is a `MultipleCaseInsensitiveDict` that we created when loading from cache. It gives us the
            # un-joined cookies, use that
            return self.response.headers.get_all(name, failobj)
        else:
            # Fall back to reading from the ', '-joined string.
            if name in self.response.headers:
                return [self.response.headers[name]]
        return failobj


KeyType = TypeVar('KeyType')

ValueType = TypeVar('ValueType')

class LookupFailure(ScraperError):
    pass

def lookup(values: Dict[KeyType, ValueType]) -> Callable[[KeyType], ValueType]:
    def _lookup(key: KeyType):
        try:
            return values[key]
        except KeyError as error:
            raise LookupFailure(repr(key)) from error
    return _lookup



#!/usr/bin/env python3

# standards
import re

# hublot
from hublot import retry_on_scraper_error


def test_basic_logging(client, server, captured_logs):
    client.get(f'{server}/hello')
    assert re.search(fr'^\[\w\w\w/\w{{13}}\]          {server}/hello\n$', captured_logs())
    client.get(f'{server}/hello')
    assert re.search(fr'^\[\w\w\w/\w{{13}}\] \[cached\] {server}/hello\n$', captured_logs())


def test_logging_courtesy_sleep(client, server, captured_logs):
    client.get(f'{server}/echo?x=1')
    assert re.search(fr'^\[\w\w\w/\w{{13}}\]          {server}/echo\?x=1\n$', captured_logs())
    client.get(f'{server}/echo?x=2')
    assert re.search(fr'^\[\w\w\w/\w{{13}}\] \[  5s  \] {server}/echo\?x=2\n$', captured_logs())


def test_logging_redirects(client, server, captured_logs):
    client.get(f'{server}/redirect/chain/1')
    assert re.search(
        r'^'
        fr'\[\w\w\w/\w{{13}}\]          {server}/redirect/chain/1\n'
        fr'\[\w\w\w/\w{{13}}\]          -> {server}/redirect/chain/2\n'
        fr'\[\w\w\w/\w{{13}}\]          -> {server}/redirect/chain/3\n'
        r'$',
        captured_logs(),
    )


def test_logging_on_retry(client, server, unique_key, captured_logs):
    @retry_on_scraper_error
    def scrape():
        client.get(f'{server}/fail-twice-then-succeed/{unique_key}')
        return 'ok'

    scrape()
    assert re.search(
        r'^'
        fr'\[\w\w\w/\w{{13}}\]          {server}/fail-twice-then-succeed/{unique_key}\n'
        'HTTPError: 500 .+ sleeping 1s\n'
        fr'\[\w\w\w/\w{{13}}\]          {server}/fail-twice-then-succeed/{unique_key}\n'
        'HTTPError: 500 .+ sleeping 5s\n'
        fr'\[\w\w\w/\w{{13}}\]          {server}/fail-twice-then-succeed/{unique_key}\n'
        r'$',
        captured_logs(),
    )



#!/usr/bin/env python3

# standards
from datetime import timedelta
from pathlib import Path
from typing import Optional, Union

# 3rd parties
from requests import PreparedRequest, Response
from requests.cookies import MockRequest

# hublot
from ..logs import LogEntry
from ..utils import MockResponse
from .key import CacheKey, UserSpecifiedCacheKey
from .storage import DiskStorage, Storage


class Cache:

    def __init__(self, storage: Storage, max_age: Optional[timedelta] = None):
        self.storage = storage
        self.max_age = max_age
        self.needs_pruned = (max_age is not None)

    def get(
        self,
        preq: PreparedRequest,
        log: LogEntry,
        key: Optional[UserSpecifiedCacheKey] = None,
        max_age: Optional[timedelta] = None,
    ) -> Optional[Response]:
        """
        Looks up the given `PreparedRequest`, and returns the corresponding `Response` if it was in cache, or `None` otherwise.
        """
        key = CacheKey.parse(key) if key else CacheKey.compute(preq)
        if self.needs_pruned and self.max_age is not None:
            self.storage.prune(self.max_age)
            self.needs_pruned = False
        if max_age is None or (self.max_age is not None and max_age > self.max_age):
            # The `max_age` passed to the method can't be greater than that given to the constructor, that would be inconsistent
            # with the cache pruning that happens and which is always based on the constructor-given `self.max_age`
            max_age = self.max_age
        res = self.storage.read(key, max_age)
        if res is not None:
            res.request = preq  # the storage doesn't need to store and recreate the request
            res.cookies.extract_cookies(MockResponse(res), MockRequest(preq))  # type: ignore[arg-type]
            log.cached = True
        else:
            res = None
            log.cached = False
        log.cache_key_str = key.unique_str
        return res

    def put(
        self,
        preq: PreparedRequest,
        log: LogEntry,
        res: Response,
        key: Optional[UserSpecifiedCacheKey] = None,
    ) -> None:
        key = CacheKey.parse(key) if key else CacheKey.compute(preq)
        log.cache_key_str = key.unique_str
        self.storage.write(key, res)

    @classmethod
    def load(
        cls,
        cache: Optional[Union['Cache', Path, str]] = None,
        max_age: Optional[timedelta] = None,
    ) -> Optional['Cache']:
        """
        Takes the `cache` param given to the `Client` constructor, and returns a `Cache` instance, or `None`
        """
        if cache is None or isinstance(cache, Cache):
            if max_age is not None:
                raise Exception("You can't specify a max_age if passing in an already instantiated `Cache`")
            return cache
        elif isinstance(cache, Path):
            return cls(
                storage=DiskStorage(root_path=cache),
                max_age=max_age,
            )
        elif isinstance(cache, str) and cache.startswith('s3://'):  # pragma: no cover
            raise NotImplementedError  # some day
        else:
            raise ValueError(repr(cache))



#!/usr/bin/env python3

# standards
from collections.abc import Generator, Iterator, Sized
from contextlib import contextmanager
from dataclasses import dataclass
from functools import wraps
import threading
from time import sleep
from typing import Callable, Optional, Sequence

# hublot
from .exceptions import ScraperError
from .logs import LOGGER


@dataclass
class ThreadLocalData:
    is_retry: bool = False


SCRAPER_LOCAL = threading.local()
SCRAPER_LOCAL.stack = [ThreadLocalData()]


def retry_on_scraper_error(
    no_parens_function: Optional[Callable] = None,
    *,
    error_types: Sequence[type] = (),
    num_attempts: int = 5,
):
    if no_parens_function:
        # decorator is without parentheses
        return retry_on_scraper_error()(no_parens_function)

    @contextmanager
    def scraper_stack_frame():
        frame = ThreadLocalData()
        SCRAPER_LOCAL.stack.append(frame)
        try:
            yield frame
        finally:
            SCRAPER_LOCAL.stack.pop()

    error_types = (ScraperError, *error_types)

    def make_wrapper(function: Callable):
        @wraps(function)
        def wrapper(*args, **kwargs):
            with scraper_stack_frame() as frame:
                for attempt in range(num_attempts):
                    try:
                        if attempt > 0:
                            frame.is_retry = True
                        payload = function(*args, **kwargs)
                        if isinstance(payload, (Iterator, Generator)) and not isinstance(payload, Sized):
                            payload = list(payload)
                        return payload
                    except Exception as error:  # pylint: disable=broad-except
                        if isinstance(error, error_types) and attempt < num_attempts - 1:
                            delay = 5 ** attempt
                            LOGGER.error('%s: %s - sleeping %ds', type(error).__name__, error, delay)
                            sleep(delay)
                        else:
                            raise
            raise Exception("can't reach here")
        return wrapper
    return make_wrapper



#!/usr/bin/env python3

# standards
import re

# hublot
from hublot import Client


def test_default_user_agent(server):
    client = Client()
    user_agent = client.get(f'{server}/echo').json()['headers']['User-Agent']
    assert re.search(r'^hublot/[\d\.]+$', user_agent), user_agent


def test_user_agent_constructor_arg(server):
    client = Client(user_agent='Nyanya/10.8')
    user_agent = client.get(f'{server}/echo').json()['headers']['User-Agent']
    assert user_agent == 'Nyanya/10.8'


def test_user_agent_instance_attribute(server):
    client = Client()
    client.user_agent='Dwidwi/0.9'
    user_agent = client.get(f'{server}/echo').json()['headers']['User-Agent']
    assert user_agent == 'Dwidwi/0.9'


def test_user_agent_in_headers(server):
    client = Client()
    res = client.get(
        f'{server}/echo',
        headers={'User-Agent': 'Bwabwa/7.3'},
    )
    user_agent = res.json()['headers']['User-Agent']
    assert user_agent == 'Bwabwa/7.3'



#!/usr/bin/env python3

# standards
from itertools import count
import json
from random import choices
from string import ascii_letters

# 3rd parties
import pytest
import requests

# hublot
from hublot import ScraperError, retry_on_scraper_error


def test_retry_decorator_no_exception(client, server):
    @retry_on_scraper_error
    def fetch():
        return client.get(f'{server}/hello').text
    assert fetch() == 'hello'


def test_retry_decorator_on_http_error(client, server, unique_key):
    @retry_on_scraper_error
    def fetch():
        return client.get(f'{server}/fail-twice-then-succeed/{unique_key}').text
    assert fetch() == 'success after 2 failures'


def test_retry_decorator_on_value_error():
    counter = count()
    @retry_on_scraper_error
    def fetch():
        i = next(counter)
        if i < 3:
            raise ValueError('x')
        return f'Success on attempt {i}'
    assert fetch() == 'Success on attempt 3'


def test_retry_decorator_num_attempts_just_enough(client, server, unique_key):
    @retry_on_scraper_error(num_attempts=3)
    def fetch():
        return client.get(f'{server}/fail-twice-then-succeed/{unique_key}').text
    assert fetch() == 'success after 2 failures'


def test_retry_decorator_num_attempts_just_not_enough(client, server, unique_key):
    @retry_on_scraper_error(num_attempts=2)
    def fetch():
        return client.get(f'{server}/fail-twice-then-succeed/{unique_key}').text
    with pytest.raises(ScraperError):
        fetch()


def test_retry_decorator_doesnt_catch_other_exceptions():
    @retry_on_scraper_error
    def fetch():
        raise KeyError('x')
    with pytest.raises(KeyError):
        fetch()


def test_retry_decorator_error_types():
    counter = count()
    @retry_on_scraper_error(error_types=[KeyError])
    def fetch():
        i = next(counter)
        if i < 3:
            raise KeyError('x')
        return f'Success on attempt {i}'
    assert fetch() == 'Success on attempt 3'


def test_if_scraper_returns_generator_it_gets_consumed():
    @retry_on_scraper_error
    def fetch():
        yield 1
        yield 2
        yield 3
    assert fetch() == [1, 2, 3]


def test_if_scraper_returns_iterator_it_gets_consumed():
    @retry_on_scraper_error
    def fetch():
        return (i for i in range(1, 4))
    assert fetch() == [1, 2, 3]


def test_if_scraper_returns_map_it_gets_consumed():
    @retry_on_scraper_error
    def fetch():
        return map(lambda x: x, range(1, 4))
    assert fetch() == [1, 2, 3]


def test_if_scraper_returns_filter_it_gets_consumed():
    @retry_on_scraper_error
    def fetch():
        return filter(lambda x: x, range(1, 4))
    assert fetch() == [1, 2, 3]


def test_if_scraper_returns_dict_keys_it_doesnt_get_consumed():
    @retry_on_scraper_error
    def fetch():
        return {1: 1, 2: 2, 3: 3}.keys()
    assert isinstance(fetch(), type({}.keys()))  # you're confused, pylint: disable=isinstance-second-argument-not-valid-type


def test_if_scraper_returns_requests_response_it_doesnt_get_turned_into_a_list():
    # requests.Response is an example of an object that is iterable, but doesn't have a __len__
    @retry_on_scraper_error
    def fetch():
        return requests.Response()
    assert isinstance(fetch(), requests.Response)


def test_scraper_sleeps_increasingly_long_delays(mocked_sleep_on_retry):
    """
    The first sleep must be >= 0 seconds, and then they must all be > than the previous
    """
    counter = count()
    @retry_on_scraper_error
    def fetch():
        i = next(counter)
        if i < 4:
            raise ValueError('x')
        return f'Success on attempt {i}'
    assert fetch() == 'Success on attempt 4'
    assert mocked_sleep_on_retry.call_count == 4
    previous_sleep = -1
    for attempt in range(4):
        called_args = mocked_sleep_on_retry.call_args_list[attempt][0]
        assert len(called_args) == 1, called_args
        sleep, = called_args
        assert sleep > previous_sleep
        previous_sleep = sleep


def test_no_courtesy_sleep_on_retries(mocked_courtesy_sleep, client, server, unique_key):
    client.get(f'{server}/hello')
    @retry_on_scraper_error
    def fetch():
        return client.get(f'{server}/fail-twice-then-succeed/{unique_key}').text
    fetch()
    sleeps = [call[0][0] for call in mocked_courtesy_sleep.call_args_list]
    # we should've slept on the 1st call b/c we'd just called the server, then no sleep on subsequent calls
    assert sleeps == [pytest.approx(5, 0.1)]


def test_decorated_function_fetches_twice(client, server):
    key_1 = ''.join(choices(ascii_letters, k=32))
    key_2 = ''.join(choices(ascii_letters, k=32))

    @retry_on_scraper_error
    def fetch():
        return [
            client.get(f'{server}/fail-twice-then-succeed/{key_1}').text,
            client.get(f'{server}/fail-twice-then-succeed/{key_2}').text,
        ]

    # The call will succeed on the 5th attempt -- the 1st call fails twice (and the 2nd isn't even reached), then the 2nd call
    # fails twice, then they both work. Because force_cache_stale=True on every retry and applies to both calls, the 1st call was
    # not cached, even after it had succeeded -- when the decorator performs a retry, *all* requests within it are uncached. For
    # that reason the 1st endpoint ends up being called 4 times, and the 2nd one 3 times.
    assert fetch() == ['success after 2 failures and 2 successes', 'success after 2 failures']


def test_decorator_on_client_from_outer_scope(reinstantiable_client, server, unique_key):
    client = reinstantiable_client()
    @retry_on_scraper_error
    def fetch():
        return client.get(f'{server}/fail-twice-then-succeed/{unique_key}').text
    assert fetch() == 'success after 2 failures'


def test_decorator_on_client_passed_as_argument(reinstantiable_client, server, unique_key):
    @retry_on_scraper_error
    def fetch(c):
        return c.get(f'{server}/fail-twice-then-succeed/{unique_key}').text
    assert fetch(reinstantiable_client()) == 'success after 2 failures'


def test_decorator_on_client_created_within_function(reinstantiable_client, server, unique_key):
    @retry_on_scraper_error
    def fetch():
        client = reinstantiable_client()
        return client.get(f'{server}/fail-twice-then-succeed/{unique_key}').text
    assert fetch() == 'success after 2 failures'


def test_parsing_html_as_json_is_a_scraper_error():
    with pytest.raises(ScraperError):
        json.loads('<html>')


def test_nested_scraper_functions(client, server, unique_key):

    # What should happen here:
    #
    #  1. the GET call in `inner()` fails, is retried, and so the call to `inner()` returns "success after 2 failures"
    #
    #  2. the GET call in `outer()` fails, and is retried; this does not force the inner call to be retried though -- that one
    #     keeps being served from cache

    @retry_on_scraper_error
    def inner():
        return 'inner: ' + client.get(f'{server}/fail-twice-then-succeed/{unique_key}-inner').text

    @retry_on_scraper_error
    def outer():
        return [
            inner(),
            'outer: ' + client.get(f'{server}/fail-twice-then-succeed/{unique_key}-outer').text,
        ]

    assert outer() == [
        'inner: success after 2 failures',
        'outer: success after 2 failures',
    ]



#!/usr/bin/env python3

# standards
from dataclasses import dataclass, replace
from hashlib import md5
import re
from typing import Optional, Tuple, Union

# 3rd parties
from requests import PreparedRequest


UserSpecifiedCacheKey = Union['CacheKey', Tuple[str, ...], str]


@dataclass(frozen=True, order=True)
class CacheKey:

    parts: Tuple[str, ...]
    sequence_num: int = 0

    @property
    def path_parts(self) -> Tuple[str, ...]:
        parts = [
            # chars that *must* be escaped:
            #  * all chars that aren't valid in Windows file names
            #  * slashes, backslashes and null chars
            #  * dots, so that ".4" at the end unambiguously identifies a sequence num (and also avoids directory traversal vulns)
            re.sub(r'[^\w\-]', lambda m: f'%{ord(m.group()):02X}', part)
            for part in self.parts
        ]
        if self.sequence_num > 0:
            parts[-1] += f'.{self.sequence_num}'
        return tuple(parts)

    @property
    def unique_str(self) -> str:
        # NB the string we return isn't for use in paths, so we can use '/' as the separator regardless of platform. Slashes have
        # been removed from the parts, so this is unambiguous.
        return '/'.join(self.path_parts)

    def next_in_sequence(self):
        return replace(self, sequence_num=self.sequence_num + 1)

    @classmethod
    def from_path_parts(cls, parts):
        seq_match = re.search(r'\.(\d+)$', parts[-1])
        if seq_match:
            sequence_num = int(seq_match.group(1))
            parts[-1] = parts[-1][:seq_match.start()]
        else:
            sequence_num = 0
        return cls(
            parts=tuple(
                re.sub(r'%([0-9a-fA-F]{2})', lambda m: chr(int(m.group(1), 16)), p)
                for p in parts
            ),
            sequence_num=sequence_num,
        )

    @classmethod
    def compute(cls, preq: PreparedRequest) -> 'CacheKey':
        # NB we don't normalise the order of the `params` dict or `data` dict. If running in Python 3.6+, where dicts preserve
        # their insertion order, multiple calls from the same code, where the params are defined in the same order, will hit the
        # same cache key. In previous versions, maybe not, so in 3.5 and before params and body should be serialised before being
        # sent to Hublot.
        headers = sorted(
            (key.title(), value)
            for key, value in preq.headers.items()
        )
        key = (
            preq.method,
            preq.url,
            headers,
            preq.body,
        )
        # Shortening to 16 chars means it's easier to copy-paste, takes less space in the terminal, etc. Seems like a flimsy reason
        # for increasing the chances of a collision, but at 2^64 bits these chances are still comfortably negligible.
        hashed = md5(repr(key).encode('UTF-8')).hexdigest()[:16]
        return cls((hashed[:3], hashed[3:]))

    @classmethod
    def parse(cls, user_specified_key: Optional[UserSpecifiedCacheKey]):
        if isinstance(user_specified_key, CacheKey):
            return user_specified_key
        elif isinstance(user_specified_key, tuple):
            return cls(user_specified_key)
        elif isinstance(user_specified_key, str):
            return cls(tuple(user_specified_key.strip('/').split('/')))
        else:
            raise TypeError(repr(user_specified_key))



#!/usr/bin/env python3

# standards
from dataclasses import dataclass
import logging
from typing import Optional, Union

# 3rd parties
from requests import PreparedRequest


LOGGER = logging.getLogger('hublot')


@dataclass(frozen=False)
class LogEntry:
    preq: PreparedRequest
    is_redirect: bool = False
    cache_key_str: Optional[str] = None
    cached: Optional[bool] = None
    courtesy_seconds: Optional[float] = None

    def _compose_line(self):
        if self.cache_key_str:
            yield f'[{self.cache_key_str}] '
        if self.cached:
            yield '[cached] '
        elif self.courtesy_seconds and self.courtesy_seconds > 0.5:
            rounded = f'{round(self.courtesy_seconds)}s'
            yield f'[{rounded:^6s}] '
        else:
            yield '         '
        if self.is_redirect:
            yield '-> '
        pr = self.preq
        yield pr.url
        if pr.method != 'GET':
            yield f' [{pr.method}'
            try:
                length = int(pr.headers.get('Content-Length', 0))
            except ValueError:
                length = 0
            if length > 0:
                yield f' {length} bytes'
            yield ']'

    def __str__(self):
        return ''.join(self._compose_line())


def basic_logging_config(level: Union[int, str] = 'INFO', propagate: bool = False):
    """
    Sets up logging for the common use case. Calls `logging.basicConfig`, lowers verbosity for the `urllib3` logger. If `propagate`
    is False (the default), a new handler will be attached to `hublot.LOGGER` that logs in a simple format to stderr, and does not
    propagate log events to the root logger.
    """
    if not isinstance(level, int):
        level = getattr(logging, level)
    logging.basicConfig(level=level)
    logging.getLogger('urllib3').setLevel(max(logging.WARNING, level))
    if not propagate and not LOGGER.handlers:
        handler = logging.StreamHandler()
        formatter = logging.Formatter('%(message)s', None, '%')
        handler.setFormatter(formatter)
        LOGGER.addHandler(handler)
        LOGGER.propagate = False



#!/usr/bin/env python3

# standards
from contextlib import contextmanager
from datetime import datetime, timedelta
import gzip
from itertools import count, product
from os import utime

# 3rd parties
import pytest


def _fetch_unique_number_then_move_clock_to_next_day(mocker, reinstantiable_client, server):
    now = datetime.now()
    client = reinstantiable_client()
    unique = client.get(f'{server}/unique-number').text
    mocker.patch('hublot.cache.storage.current_datetime', return_value=now + timedelta(days=1))
    return unique


def test_cache_max_age_defaults(mocker, reinstantiable_client, server):
    # no max_age specified, cache is still valid a day later
    unique = _fetch_unique_number_then_move_clock_to_next_day(mocker, reinstantiable_client, server)
    client = reinstantiable_client()
    assert unique == client.get(f'{server}/unique-number').text


def test_cache_long_max_age(mocker, reinstantiable_client, server):
    # max_age is 2 days, cache is still valid a day later
    unique = _fetch_unique_number_then_move_clock_to_next_day(mocker, reinstantiable_client, server)
    client = reinstantiable_client()
    assert unique == client.get(f'{server}/unique-number', max_cache_age=timedelta(days=2)).text


def test_cache_short_max_age(mocker, reinstantiable_client, server):
    # max_age is 12 hours, cache is no longer valid a day later, we get a new `unique` number
    unique = _fetch_unique_number_then_move_clock_to_next_day(mocker, reinstantiable_client, server)
    client = reinstantiable_client()
    assert unique != client.get(f'{server}/unique-number', max_cache_age=timedelta(hours=12)).text


def test_cache_pruning(mocker, reinstantiable_client, server):
    # client is created with max_age of 12 hours, and a request (of any URL) is run. The cache gets pruned.
    unique = _fetch_unique_number_then_move_clock_to_next_day(mocker, reinstantiable_client, server)
    client = reinstantiable_client(max_age=timedelta(hours=12))
    client.get(f'{server}/hello')
    # now it's no longer cached, we fetch a new `unique` number
    client = reinstantiable_client()
    assert unique != client.get(f'{server}/unique-number').text


def test_override_with_method_kwarg(mocker, reinstantiable_client, server):
    # instantiate cache with a max_age that would accept the cached file, but then call the get() method with a shorter age -- the
    # method kwarg overrides the constructor kwarg, cache is invalidated, a new value is fetched
    unique = _fetch_unique_number_then_move_clock_to_next_day(mocker, reinstantiable_client, server)
    client = reinstantiable_client(max_age=timedelta(days=10))
    assert unique != client.get(f'{server}/unique-number', max_cache_age=timedelta(hours=12)).text


def test_cant_override_with_longer_age(mocker, reinstantiable_client, server):
    # instantiate cache with a max_age that doesn't accept the cached file. Try overriding it with a longer max_cache_age -- that
    # doesn't work. The constructor-given max_cache_age is an upper bound on what the method kwarg accepts (to stay consistent with
    # the fact that the constructor-given max_cache_age triggers full prunes)
    now = datetime.now()
    client = reinstantiable_client(max_age=timedelta(hours=12))
    unique = client.get(f'{server}/unique-number').text
    mocker.patch('hublot.cache.storage.current_datetime', return_value=now + timedelta(days=1))
    # re-use the same client to ensure there's no pruning happening, that would render the test invalid
    assert unique != client.get(f'{server}/unique-number', max_cache_age=timedelta(days=2)).text


def test_cant_override_with_null_age(mocker, reinstantiable_client, server):
    # instantiate cache with a max_age that doesn't accept the cached file. Try overriding it with max_cache_age=None -- that also
    # doesn't work, we fall back to the constructor-given kwarg
    now = datetime.now()
    client = reinstantiable_client(max_age=timedelta(hours=12))
    unique = client.get(f'{server}/unique-number').text
    mocker.patch('hublot.cache.storage.current_datetime', return_value=now + timedelta(days=1))
    # again, re-use the same client to ensure there's no pruning
    assert unique != client.get(f'{server}/unique-number', max_cache_age=None).text


@pytest.mark.parametrize(
    'redirect_step_is_cached',
    product([False, True], repeat=3),
)
def test_max_age_applies_when_following_redirects(mocker, reinstantiable_client, server, redirect_step_is_cached):
    now = datetime.now()
    mocker.patch('hublot.cache.storage.current_datetime', lambda: now)

    @contextmanager
    def mocked_gzip_open(path, *rest, **kwargs):
        with _real_gzip_open(path, *rest, **kwargs) as f:
            yield f
        utime(path, (now.timestamp(), now.timestamp()))
    _real_gzip_open = gzip.open
    mocker.patch('hublot.cache.storage.gzip.open', mocked_gzip_open)

    client = reinstantiable_client(
        cookies_enabled=False,  # disable cookies as they would invalidate the cache
    )

    # first, cache every step of the redirect chain
    client.fetch(f'{server}/redirect/chain/1')

    # then on the next day, freshen the cache for the redirect steps that should be read from cache
    now += timedelta(hours=24)
    for step_num, is_cached in zip(count(1), redirect_step_is_cached):
        if is_cached:
            client.fetch(
                f'{server}/redirect/chain/{step_num}',
                allow_redirects=False,
                force_cache_stale=True,
            )

    # now if you re-fetch the whole chain, only the ones that were freshly fetched should be read from cache
    now += timedelta(hours=12)
    res = client.fetch(
        f'{server}/redirect/chain/1',
        max_cache_age=timedelta(hours=24),
    )

    assert [r.from_cache for r in [*res.history, res]] == list(redirect_step_is_cached)



#!/usr/bin/env python3

# standards
from collections import OrderedDict
from itertools import count, product

# 3rd parties
import pytest
from requests import Request

# hublot
from hublot.cache import CacheKey
from hublot.logs import LogEntry
from .utils import assert_responses_equal, dummy_prepared_request, dummy_response, iter_equal_pairs, iter_nonequal_pairs


def test_simple_cache_use(client):
    preq = dummy_prepared_request(client)
    log = LogEntry(preq)
    cache = client.cache
    assert cache.get(preq, log) is None
    assert log.cache_key_str is not None
    cache.put(preq, log, dummy_response(preq))
    assert_responses_equal(cache.get(preq, log), dummy_response(preq))


EQUIVALENCIES = [
    # For every attribute of a PreparedRequest, lists groups values such that values within the same group should be cached under
    # the same key, but not across groups

    # Obviously requests using different methods should be cached separately
    [{'method': 'GET'}],
    [{'method': 'POST'}],

    # URLs are case sensitive
    [{'url': 'http://cache-test/url'}],
    [{'url': 'http://cache-test/URL'}],
    [{'url': 'http://cache-test/URL/'}],

    # The ordering of the parameters in the URL string matters. They're not parsed at all, so a final '&' is also a cache breaker
    [{'url': 'http://cache-test/TEST?a=1&b=2'}],
    [{'url': 'http://cache-test/TEST?b=2&a=1'}],
    [{'url': 'http://cache-test/TEST?a=1&b=2&'}],

    # `params` are also part of the URL key
    [{'url': 'http://cache-test/params', 'params': {}}],
    [{'url': 'http://cache-test/params', 'params': {'x': 'a'}}],

    # `params` get appended to the URL, so these are equivalent. Using OrderedDict for pythons before 3.7.
    [
        {'url': 'http://cache-test/params-test', 'params': OrderedDict([('a', '1'), ('b', '2')])},
        {'url': 'http://cache-test/params-test?', 'params': OrderedDict([('a', '1'), ('b', '2')])},
        {'url': 'http://cache-test/params-test?a=1', 'params': OrderedDict([('b', '2')])},
        {'url': 'http://cache-test/params-test?a=1&b=2', 'params': OrderedDict()},
    ],

    # The presence, absence, or value of a header are all enough to bust the cache
    [{'url': 'http://cache-test/header-test', 'headers': {}}],
    [{'url': 'http://cache-test/header-test', 'headers': {'X-Test': '1'}}],
    [{'url': 'http://cache-test/header-test', 'headers': {'X-Test': '2'}}],

    # The `requests` library will normalise the case of headers before sending them off to the server, and so headers are
    # case-insensitive, and so all of these get the same cache key
    [
        {'url': 'http://cache-test/header-test-2', 'headers': {'X-Test': '1'}},
        {'url': 'http://cache-test/header-test-2', 'headers': {'x-test': '1'}},
        {'url': 'http://cache-test/header-test-2', 'headers': {'X-TEST': '1'}},
    ],

    # The order of the headers in the dict doesn't matter of course
    [
        {'url': 'http://cache-test/header-order-test', 'headers': {'X-Test-A': 'a', 'X-Test-B': 'b'}},
        {'url': 'http://cache-test/header-order-test', 'headers': {'X-Test-B': 'b', 'X-Test-A': 'a'}},
    ],

    # Setting a cookie via `cookies` is the same as setting it manually in a header
    [
        {'url': 'http://cache-test/cookie-test', 'cookies': {'a': '1'}, 'headers': {}},
        {'url': 'http://cache-test/cookie-test', 'cookies': {}, 'headers': {'Cookie': 'a=1'}},
    ],

    # Setting `data` as a dict is equivalent to setting it manually as bytes
    [
        {'url': 'http://cache-test/data-test', 'data': {'a': '1'}},
        {
            'url': 'http://cache-test/data-test',
            'data': b'a=1',
            'headers': {'Content-Type': 'application/x-www-form-urlencoded', 'Content-Length': '3'},
        },
    ],

    # Setting `json` is equivalent to manually building the request
    [
        {'url': 'http://cache-test/json-test', 'json': {'a': '1'}},
        {
            'url': 'http://cache-test/json-test',
            'data': b'{"a": "1"}',
            'headers': {'Content-Type': 'application/json'},
        },
    ],

    # requests can be specified as `Request` objects
    [
        {'url': 'http://cache-test/request-objects', 'method': 'GET'},
        {'url': 'http://cache-test/request-objects', 'method': 'GET', 'params': {}},
        {'url': Request(url='http://cache-test/request-objects', method='GET')},
        {'url': Request(url='http://cache-test/request-objects', method='GET', params={})},
    ],
    [
        {'url': 'http://cache-test/request-objects', 'method': 'POST', 'data': {}},
        {'url': Request(url='http://cache-test/request-objects', method='POST', data={})},
    ],
    [
        {
            'url': 'http://cache-test/request-objects',
            'method': 'POST',
            'data': {'a': '1'},
        },
        {
            'url': 'http://cache-test/request-objects',
            'method': 'POST',
            'data': 'a=1',
            'headers': {'Content-Type': 'application/x-www-form-urlencoded'},
        },
        {
            'url': Request(
                url='http://cache-test/request-objects',
                method='POST',
                data={'a': '1'},
            ),
        },
        {
            'url': Request(
                url='http://cache-test/request-objects',
                method='POST',
                data='a=1',
                headers={'Content-Type': 'application/x-www-form-urlencoded'},
            ),
        },
    ],

]


@pytest.mark.parametrize(
    'config_1, config_2',
    iter_nonequal_pairs(EQUIVALENCIES),
)
def test_unique_keys(client, config_1, config_2):
    key = CacheKey.compute(dummy_prepared_request(client, **config_1))
    other_key = CacheKey.compute(dummy_prepared_request(client, **config_2))
    assert key != other_key


@pytest.mark.parametrize(
    'config_1, config_2',
    iter_nonequal_pairs(EQUIVALENCIES),
)
def test_unique_requests(client, config_1, config_2):
    cache = client.cache
    preq_1 = dummy_prepared_request(client, **config_1)
    preq_2 = dummy_prepared_request(client, **config_2)
    cache.put(preq_1, LogEntry(preq_1), dummy_response(preq_1))
    assert cache.get(preq_2, LogEntry(preq_2)) is None


@pytest.mark.parametrize(
    'config_1, config_2',
    iter_equal_pairs(EQUIVALENCIES),
)
def test_equivalent_keys(client, config_1, config_2):
    key_1 = CacheKey.compute(dummy_prepared_request(client, **config_1))
    key_2 = CacheKey.compute(dummy_prepared_request(client, **config_2))
    print(dummy_prepared_request(client, **config_1).__dict__)
    print(dummy_prepared_request(client, **config_2).__dict__)
    assert key_1 == key_2, (config_1, config_2)


@pytest.mark.parametrize(
    'config_1, config_2',
    iter_equal_pairs(EQUIVALENCIES),
)
def test_equivalent_requests(client, config_1, config_2):
    cache = client.cache
    preq_1 = dummy_prepared_request(client, **config_1)
    preq_2 = dummy_prepared_request(client, **config_2)
    response = dummy_response(preq_1)
    assert cache.get(preq_2, LogEntry(preq_2)) is None  # else test is invalid
    cache.put(preq_1, LogEntry(preq_1), response)
    assert_responses_equal(
        cache.get(preq_2, LogEntry(preq_2)),
        response,
    )


def test_cache_updates_log_entry_attributes(client):
    cache = client.cache
    preq = dummy_prepared_request(client)
    log = LogEntry(preq)
    assert log.cache_key_str is None
    assert log.cached is None
    cache.get(preq, log)
    assert log.cache_key_str is not None
    assert log.cached is False
    cache.put(preq, log, dummy_response(preq))
    cache.get(preq, log)
    assert log.cached is True


TEST_KEYS = [
    ('simple-string', '/simple-string', 'simple-string'),
    ('space string', '/space%20string', 'space%20string'),
    ('slash/string', '/slash/string', 'slash/string'),
    ('/slash/string', '/slash/string', 'slash/string'),
    ('slash/string/', '/slash/string', 'slash/string'),
    (('item', '123'), '/item/123', 'item/123'),
    (('slash', '/'), '/slash/%2F', 'slash/%2F'),
]


@pytest.mark.parametrize(
    'user_specified, expected_path, expected_unique_str',
    TEST_KEYS
)
def test_cache_key_parsing(user_specified, expected_path, expected_unique_str):
    parsed = CacheKey.parse(user_specified)
    assert ''.join(f'/{p}' for p in parsed.path_parts) == expected_path
    assert parsed.unique_str == expected_unique_str


@pytest.mark.parametrize(
    'user_specified',
    [spec[0] for spec in TEST_KEYS]
)
def test_cache_key_parsing_from_path_parts(user_specified):
    parsed = CacheKey.parse(user_specified)
    assert CacheKey.from_path_parts(parsed.path_parts) == parsed


def test_user_specified_cache_key(client, server):
    counter = count()
    all_keys = ['one', 'two', 'three']
    all_values = [
        client.get(
            f'{server}/unique-number',
            cache_key=key,
            params={'unique': str(next(counter))}
        ).text
        for key in all_keys
    ]
    assert len(set(all_values)) == len(all_values)  # they're all different
    for key, expected in zip(all_keys, all_values):
        obtained = client.get(
            f'{server}/unique-number',
            cache_key=key,
            # the param is actually different, but the cache key isn't, so we should get the same value back
            params={'unique': str(next(counter))}
        ).text
        assert obtained == expected


def test_user_specified_cache_key_on_redirect(client, server):
    res = client.get(
        f'{server}/redirect/chain/1',
        cache_key='fixed',
    )
    assert res.text == 'Landed'
    all_keys_in_cache = sorted(k.unique_str for k in client.cache.storage.iter_all_keys())
    assert all_keys_in_cache == ['fixed', 'fixed.1', 'fixed.2']


@pytest.mark.parametrize(
    'key_1, key_2, should_match',
    [
        (key_1, key_2, group_1 is group_2)
        for group_1, group_2 in product(
            # Keys from the same line here should match each other, and keys from different lines shouldn't
            [
                (CacheKey(parts=('a', 'b')), ('a', 'b'), 'a/b'),
                (CacheKey(parts=('A', 'B')), ('A', 'B'), 'A/B'),
                (CacheKey(parts=('b', 'c')), ('b', 'c'), 'b/c'),
            ],
            repeat=2,
        )
        for key_1 in group_1
        for key_2 in group_2
    ]
)
def test_different_ways_to_express_cache_keys(client, server, key_1, key_2, should_match):
    counter = count()
    response_1, response_2 = (
        client.get(
            f'{server}/unique-number',
            cache_key=key,
            params={'unique': str(next(counter))}
        ).text
        for key in (key_1, key_2)
    )
    if should_match:
        assert response_1 == response_2
    else:
        assert response_1 != response_2


def test_cache_key_of_unknown_class(client):
    class MyRandoClass:
        pass

    with pytest.raises(TypeError):
        client.fetch(
            'http://whatever/',
            cache_key=MyRandoClass(),
        )



#!/usr/bin/env python3

"""
Functions for serialising request and response objects to binary blobs, which can then be stored in any storage.

The format used is designed to mimic the HTTP exchange that travels on the wire, i.e. the blob starts with an loose rendering of
the HTTP request head, then request body if present, then the response head, then body. This makes it convenient to inspect the
blobs for manual debugging.
"""

# standards
from codecs import getwriter
from io import BytesIO
import re
from typing import Callable, Optional, Tuple

# 3rd parties
from requests import PreparedRequest, Response
from requests.structures import CaseInsensitiveDict


EOL = '\r\n'
EOL_BYTES = EOL.encode('UTF-8')
EOL_LEN = len(EOL)


Writer = getwriter('UTF-8')  # pylint: disable=invalid-name


def compose_binary_blob(res: Response) -> bytes:
    output = BytesIO()
    write = Writer(output).write
    _compose_request_blob(res.request, output, write)
    _compose_response_blob(res, output, write)
    return output.getvalue()


def _compose_request_blob(
    preq: PreparedRequest,
    output: BytesIO,
    write: Callable[[str], None],
) -> None:

    write(f'{preq.method} {preq.url}{EOL}')
    for key, value in sorted(preq.headers.items()):
        write(f'{key}: {value}{EOL}')
    write(EOL)
    if 'Content-Length' in preq.headers:
        body = b'' if preq.body is None else preq.body
        content_length = int(preq.headers['Content-Length'])
        if content_length != len(body):
            # Don't write it out because we won't be able to read it back
            raise Exception(f'body has {len(body)} bytes but Content-Length is {content_length}')
        if isinstance(body, str):
            # `PreparedRequest.prepare_body` leaves `body` as a `str` if `data` was a dict
            body_bytes = body.encode('UTF-8')
        else:
            body_bytes = body
        output.write(body_bytes)
        write(EOL)
    write(EOL)


def _compose_response_blob(
    res: Response,
    output: BytesIO,
    write: Callable[[str], None],
) -> None:

    write(f'HTTP {res.status_code} {res.reason}{EOL}')
    if hasattr(res.raw, '_fp'):
        # Using `res.raw._fp.headers` rather than `res.headers` means we store repeated headers (e.g. Set-Cookie) as separate lines
        header_items = res.raw._fp.headers.items()  # pylint: disable=protected-access
    else:
        # Sometimes however, in tests especially, `raw` might've been replaced by some other file object, and has no `_fp`, so fall
        # back to this:
        header_items = res.headers.items()
    for key, value in sorted(header_items):
        write(f'{key}: {value}{EOL}')
    write(EOL)
    if res.content is not None:
        output.write(res.content)


def parse_binary_blob(data: bytes) -> Response:
    # pylint: disable=protected-access
    pos = 0
    _method_unused, url, pos = _parse_line(data, pos, r'^(\w+) (.+)$')
    _headers_unused, _body_unused, pos = _parse_message(data, pos)
    status_code, reason, pos = _parse_line(data, pos, r'^HTTP (\d+) (.*)$')
    res = Response()
    res.status_code = int(status_code)
    res.reason = reason
    res.url = url
    res.headers, res._content, pos = _parse_message(data, pos, read_to_end=True)
    res._content_consumed = True  # type: ignore[attr-defined]
    return res


def _parse_message(
    data: bytes,
    pos: int,
    read_to_end: bool = False,
) -> Tuple[CaseInsensitiveDict, Optional[bytes], int]:

    headers = MultipleCaseInsensitiveDict()
    while data[pos : pos+EOL_LEN] != EOL_BYTES:
        key, value, pos = _parse_line(data, pos, r'^([^:]+): (.*)$')
        headers[key] = value
    pos += EOL_LEN
    body: Optional[bytes] = None
    if read_to_end:
        body = data[pos : ]
        pos = len(data)
    elif headers.get('Content-Length'):
        length = int(headers['Content-Length'])
        body = data[pos : pos+length]
        pos += length + (2 * EOL_LEN)
    else:
        pos += EOL_LEN
    return headers, body, pos


def _parse_line(data: bytes, pos: int, regex: str):
    eol_pos = data.find(EOL_BYTES, pos)
    line = data[pos : eol_pos].decode('UTF-8')
    match = re.search(regex, line)
    if not match:
        raise ValueError(repr(line))
    return (*match.groups(), eol_pos + EOL_LEN)



class MultipleCaseInsensitiveDict(CaseInsensitiveDict):  # pylint: disable=too-many-ancestors

    def __setitem__(self, key, item):
        values = self.get_all(key, [])
        values.append(item)
        super().__setitem__(key, values)

    def __getitem__(self, key):
        values = super().__getitem__(key)
        return ', '.join(values)

    def get_all(self, key, failobj=None):
        # Deliberately copying the interface of `email.message.EmailMessage.get_all` for consistency's sake, though we're not
        # actually using this object in place of that.
        try:
            return super().__getitem__(key)
        except KeyError:
            return failobj



#!/usr/bin/env python3

# standards
from datetime import timedelta
from pathlib import Path
from tempfile import TemporaryDirectory

# 3rd parties
import pytest
import requests

# hublot
from hublot import Cache, Client, CourtesySleep
from hublot.cache.storage import DiskStorage


@pytest.mark.parametrize(
    'kwargs, expected_method',
    [
        ({}, 'GET'),
        ({'data': None}, 'GET'),
        ({'data': b''}, 'POST'),
        ({'data': b'a=b'}, 'POST'),
        ({'data': {'a': 'b'}}, 'POST'),
        ({'json': None}, 'GET'),
        ({'json': ''}, 'POST'),
        ({'json': {'a': 'b'}}, 'POST'),
    ]
)
def test_default_method(client, server, kwargs, expected_method):
    assert client.fetch(f'{server}/method-test', **kwargs).text == expected_method


def test_no_cache_by_default(server):
    client = Client()
    one = client.get(f'{server}/unique-number').text
    two = client.get(f'{server}/unique-number').text
    assert one != two  # not cached


def test_null_cache(server):
    client = Client(cache=None)
    one = client.get(f'{server}/unique-number').text
    two = client.get(f'{server}/unique-number').text
    assert one != two  # not cached


def test_cache_as_path(server):
    with TemporaryDirectory() as tmp:
        client = Client(cache=Path(tmp))
        one = client.get(f'{server}/unique-number').text
        two = client.get(f'{server}/unique-number').text
    assert one == two  # cached


def test_cache_as_cache_object(server):
    with TemporaryDirectory() as tmp:
        client = Client(cache=Cache(DiskStorage(Path(tmp))))
        one = client.get(f'{server}/unique-number').text
        two = client.get(f'{server}/unique-number').text
    assert one == two  # cached


def test_force_cache_stale(client, server):
    one = client.get(f'{server}/unique-number').text
    two = client.get(f'{server}/unique-number', force_cache_stale=True).text
    three = client.get(f'{server}/unique-number').text
    assert one != two  # cache not read on 2nd call
    assert two == three  # but cache was written on 2nd call


def test_courtesy_sleep_by_default(mocked_courtesy_sleep, server):
    client = Client()
    client.get(f'{server}/unique-number')
    client.get(f'{server}/unique-number')
    mocked_courtesy_sleep.assert_called_once()
    delay, = mocked_courtesy_sleep.call_args[0]
    assert delay > 1


def test_null_courtesy_sleep(mocked_courtesy_sleep, server):
    client = Client(courtesy_sleep=None)
    client.get(f'{server}/unique-number')
    client.get(f'{server}/unique-number')
    mocked_courtesy_sleep.assert_not_called()


def test_courtesy_sleep_as_timedelta(mocked_courtesy_sleep, server):
    client = Client(courtesy_sleep=timedelta(minutes=2))
    client.get(f'{server}/unique-number')
    client.get(f'{server}/unique-number')
    mocked_courtesy_sleep.assert_called_once()
    delay, = mocked_courtesy_sleep.call_args[0]
    assert delay == pytest.approx(120, 0.1)


def test_courtesy_sleep_as_object(mocked_courtesy_sleep, server):
    client = Client(courtesy_sleep=CourtesySleep(timedelta(seconds=78)))
    client.get(f'{server}/unique-number')
    client.get(f'{server}/unique-number')
    mocked_courtesy_sleep.assert_called_once()
    delay, = mocked_courtesy_sleep.call_args[0]
    assert delay == pytest.approx(78, 0.1)


def test_post_data(client, server):
    res = client.post(f'{server}/echo', data={'a': 'b'})
    payload = res.json()
    payload.pop('headers')
    assert payload == {'method': 'POST', 'args': {}, 'files': {}, 'form': {'a':'b'}, 'json': None}


def test_post_open_file(client, server):
    dummy_file = Path(__file__)
    res = client.post(f'{server}/echo', data=dummy_file.open('rb'))
    payload = res.json()
    payload.pop('headers')
    assert payload == {
        'method': 'POST',
        'args': {},
        'files': {},
        'form': dummy_file.read_text('UTF-8'),
        'json': None,
    }


def test_post_open_file_as_form_field(client, server):
    dummy_file = Path(__file__)
    res = client.post(f'{server}/echo', data={'a': dummy_file.open('rb')})
    payload = res.json()
    payload.pop('headers')
    assert payload == {
        'method': 'POST',
        'args': {},
        'files': {},
        'form': {'a': dummy_file.read_text('UTF-8')},
        'json': None,
    }


def test_post_files(client, server):
    dummy_file = Path(__file__)
    res = client.post(f'{server}/echo', files={'f': dummy_file.open('rb')})
    payload = res.json()
    payload.pop('headers')
    assert payload == {
        'method': 'POST',
        'args': {},
        'files': {'f': dummy_file.read_text('UTF-8')},
        'form': {},
        'json': None,
    }


def test_post_json(client, server):
    res = client.post(f'{server}/echo', json={'a': 'b'})
    payload = res.json()
    payload.pop('headers')
    assert payload == {'method': 'POST', 'args': {}, 'files': {}, 'form': {}, 'json': {'a': 'b'}}


def test_http_errors_are_raised(client, server):
    with pytest.raises(requests.HTTPError):
        client.get(f'{server}/fail-with-random-value')


def test_auto_raise_can_be_disabled(client, server):
    res = client.get(f'{server}/fail-with-random-value', raise_for_status=False)
    assert res.status_code == 500


def test_redirect(client, server):
    res = client.get(f'{server}/redirect/chain/1')
    assert res.status_code == 200
    assert res.text == 'Landed'


def test_no_redirect(client, server):
    res = client.get(f'{server}/redirect/chain/1', allow_redirects=False)
    assert res.status_code == 302
    assert res.text == 'Bounce 1'


def test_redirect_response_bodies(cache, server):
    for _ in (1, 2):
        client = Client(cache=cache)
        res = client.get(f'{server}/redirect/chain/1')
        assert res.status_code == 200
        assert res.text == 'Landed'


def test_redirects_set_response_history(cache, server):
    for _ in (1, 2):
        client = Client(cache=cache)
        res = client.get(f'{server}/redirect/chain/1')
        assert [r.text for r in res.history] == ['Bounce 1', 'Bounce 2']


def test_redirect_loop(client, server):
    # Make sure that the caching doesn't interfere with Requests' ability to detect redirect loops
    with pytest.raises(requests.TooManyRedirects):
        client.fetch(f'{server}/redirect/loop')


def test_client_preserves_casing_of_percent_escapes_in_path(client, server):
    ref_upper = client.get(f'{server}/bicam%C3%A9ral').text
    assert ref_upper.startswith('upper')
    ref_lower = client.get(f'{server}/bicam%c3%a9ral').text
    assert ref_lower.startswith('lower')
    assert client.get(f'{server}/bicam%C3%A9ral').text == ref_upper
    assert client.get(f'{server}/bicam%c3%a9ral').text == ref_lower


def test_client_preserves_casing_of_percent_escapes_in_query(client, server):
    ref_upper = client.get(f'{server}/bicam%C3%A9ral?name=Zo%C3%A9').text
    assert ref_upper.startswith('upper')
    ref_lower = client.get(f'{server}/bicam%C3%A9ral?name=Zo%c3%a9').text
    assert ref_lower.startswith('lower')
    assert client.get(f'{server}/bicam%C3%A9ral?name=Zo%C3%A9').text == ref_upper
    assert client.get(f'{server}/bicam%C3%A9ral?name=Zo%c3%a9').text == ref_lower


def test_client_can_fetch_from_server_that_redirects_based_on_escape_code_case(client, server):
    url = f'{server}/redirig%C3%A9'
    with pytest.raises(requests.TooManyRedirects):
        requests.get(url)  # doesn't work with `requests`
    assert client.get(url).text == 'lower'  # Hublot can get around it though



#!/usr/bin/env python3

# It's the pytest way, pylint: disable=redefined-outer-name

# standards
from io import StringIO
from itertools import count
import logging
from pathlib import Path
from random import choices, random, randrange
from string import ascii_letters
from tempfile import TemporaryDirectory
from threading import Thread

# 3rd parties
from flask import Flask, jsonify, make_response, request
import pytest
from werkzeug.serving import make_server  # installed transitively by Flask

# hublot
from hublot import Cache, Client, basic_logging_config, logger
import hublot.courtesy
import hublot.decorator


basic_logging_config(level='DEBUG')


@pytest.fixture
def reinstantiable_cache():
    """
    A callable that can be called repeatedly to reinstantiate the same cache. The idea is to test what happens if you discard a
    cache object then re-create it, with the same parameters, as happens when you re-run a script.
    """
    with TemporaryDirectory() as temp_root:
        yield lambda **kwargs: Cache.load(Path(temp_root), **kwargs)


@pytest.fixture
def reinstantiable_client(reinstantiable_cache):
    """
    A callable that can be called repeatedly to reinstantiate a Client with the same cache parameters.
    """
    yield lambda cookies_enabled=True, **kwargs: Client(
        cache=reinstantiable_cache(**kwargs),
        cookies_enabled=cookies_enabled,
    )


@pytest.fixture
def cache(reinstantiable_cache):
    yield reinstantiable_cache()


@pytest.fixture
def client(cache):
    yield Client(cache=cache)


def flask_app():
    # Yeah, we don't call these directly, but they still need names, pylint: disable=unused-variable
    # And yes, it's a bit long for a function, but it's still readable, pylint: disable=too-many-locals
    app = Flask('hublot-tests')

    @app.route('/hello')
    def hello():
        return 'hello'

    @app.route('/method-test', methods=['GET', 'POST', 'PUT'])
    def method_test():
        return request.method

    iter_numbers = count()
    @app.route('/unique-number', methods=['GET', 'POST'])
    def unique_number():
        return str(next(iter_numbers))

    @app.route('/echo', methods=['GET', 'POST'])
    def echo():
        return jsonify({
            'method': request.method,
            'args': request.args,
            'files': {
                key: storage.read().decode('UTF-8')
                for key, storage in request.files.items()
            },
            'form': request.data.decode('UTF-8') if request.data and not request.json else request.form,
            'json': request.json,
            'headers': dict(request.headers.items()),
        })

    @app.route('/fail-with-random-value')
    def fail_with_random_value():
        return str(random()), 500

    num_calls_by_key = {}
    num_failures_by_key = {}
    @app.route('/fail-twice-then-succeed/<key>')
    def fail_twice_then_succeed(key):
        num_calls = num_calls_by_key.get(key, 0)
        num_calls_by_key[key] = num_calls + 1
        num_failures = num_failures_by_key.get(key, 0)
        if num_calls < 2:
            num_failures_by_key[key] = num_failures + 1
            return f'crash {num_failures}', 500
        status = f'success after {num_failures} failures'
        if num_calls > num_failures:
            status += f' and {num_calls - num_failures} successes'
        return status, 200

    @app.route('/cookies/get')
    def get_cookie():
        return request.cookies

    @app.route('/cookies/set')
    def set_cookie():
        res = make_response()
        for key, value in request.args.items():
            res.set_cookie(key, value)
        return res

    @app.route('/cookies/set-two-cookies')
    def set_two_cookies():
        return str(next(iter_numbers)), 200, {'Set-Cookie': ['a=1', 'b=2']}

    @app.route('/redirect/chain/1')
    def redirect_chain_1():
        res = make_response('Bounce 1', 302)
        res.headers['Location'] = '/redirect/chain/2'
        res.set_cookie('redirect1', 'yes')
        return res

    @app.route('/redirect/chain/2')
    def redirect_chain_2():
        res = make_response('Bounce 2', 302)
        res.headers['Location'] = '/redirect/chain/3'
        res.set_cookie('redirect2', 'yes')
        return res

    @app.route('/redirect/chain/3')
    def redirect_chain_3():
        res = make_response('Landed')
        res.set_cookie('redirect3', 'yes')
        return res

    @app.route('/redirect/loop')
    def redirect_loop():
        res = make_response('Loop 1', 302)
        res.headers['Location'] = '/redirect/loop-back'
        return res

    @app.route('/redirect/loop-back')
    def redirect_loop_back():
        res = make_response('Loop 2', 302)
        res.headers['Location'] = '/redirect/loop'
        return res

    @app.route('/no-reason')
    def no_reason():
        return 'hello', '200 '  # <-- no 'reason' message after the '200'

    @app.route('/self-redirect')
    def self_redirect():
        key = 'thisismycookie'
        if request.cookies.get(key):
            return 'ok'
        else:
            res = make_response('', 302)
            res.headers['Location'] = '/self-redirect'
            res.set_cookie(key, 'ok')
            return res

    @app.route('/bicaméral')
    def bicameral():
        raw_uri = request.environ['RAW_URI']
        case = {
            '/bicam%C3%A9ral': 'upper',
            '/bicam%c3%a9ral': 'lower',
            '/bicam%C3%A9ral?name=Zo%C3%A9': 'upper',
            '/bicam%C3%A9ral?name=Zo%c3%a9': 'lower',
        }[raw_uri]
        return '%s[%s]' % (case, next(iter_numbers))

    @app.route('/redirigé')
    def redirige():
        raw_uri = request.environ['RAW_URI']
        case = {
            '/redirig%C3%A9': 'upper',
            '/redirig%c3%a9': 'lower',
        }[raw_uri]
        if case == 'upper':
            res = make_response('Zzzwip', 302)
            res.headers['Location'] = '/redirig%c3%a9'
            return res
        return 'lower'

    return app


@pytest.fixture(scope='session')
def server():
    app = flask_app()
    port = randrange(5000, 50000)
    server = make_server('127.0.0.1', port, app)  # pylint: disable=redefined-outer-name
    app.app_context().push()
    thread = Thread(target=server.serve_forever)
    thread.start()
    try:
        yield f'http://127.0.0.1:{port}'
    finally:
        server.shutdown()
        thread.join()


@pytest.fixture
def mocked_courtesy_sleep(mocker):
    mocker.patch('hublot.courtesy.sleep')
    return hublot.courtesy.sleep


@pytest.fixture
def mocked_sleep_on_retry(mocker):
    mocker.patch('hublot.decorator.sleep')
    return hublot.decorator.sleep


@pytest.fixture
def unique_key():
    return ''.join(choices(ascii_letters, k=32))


@pytest.fixture
def captured_logs():
    handler = logging.StreamHandler(StringIO())
    original_handlers = logger.handlers
    logger.handlers = [handler]

    def getvalue():
        value = handler.stream.getvalue()
        handler.stream = StringIO()
        logging.debug('Captured logs: %r', value)
        return value
    yield getvalue

    logger.handlers = original_handlers


def pytest_collection_modifyitems(items):
    for item in items:
        # these are always applied, whether or not the tests want them
        item.fixturenames.append('mocked_courtesy_sleep')
        item.fixturenames.append('mocked_sleep_on_retry')



#!/usr/bin/env python3

# standards
from pathlib import Path
import re

# 3rd parties
import setuptools


version_file = Path(__file__).parent / 'hublot' / 'version.py'
version_match = re.search(
    r"HUBLOT_VERSION = \'(.+)\'",
    version_file.read_text('UTF-8'),
)
if not version_match:
    raise Exception("Couldn't parse version.py")
HUBLOT_VERSION = version_match.group(1)


setuptools.setup(
    name='hublot',
    version=HUBLOT_VERSION,
    author='Hervé Saint-Amand',
    packages=setuptools.find_packages(),
    package_data={'hublot': ['py.typed']},
    install_requires=[
        'dataclasses>=0.8,<1; python_version<"3.7"',
        'requests>=2.25,<3',
    ],
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
    ],
    zip_safe=False, # https://mypy.readthedocs.io/en/latest/installed_packages.html#creating-pep-561-compatible-packages
)



#!/usr/bin/env python3

# You're confused because we're patching the `sleep` function and it's not what you expect, pylint: disable=no-member

# standards
from datetime import timedelta
from itertools import combinations

# 3rd parties
import pytest

# hublot
from hublot import Client
from .utils import dummy_prepared_request, dummy_response


@pytest.mark.parametrize(
    'courtesy_sleep',
    [None, timedelta(seconds=0), timedelta(seconds=5), timedelta(seconds=37)],
)
def test_courtesy_sleep(mocked_courtesy_sleep, server, courtesy_sleep):
    kwargs = {} if courtesy_sleep is None else {'courtesy_sleep': courtesy_sleep}
    client = Client(**kwargs)
    client.fetch(f'{server}/hello')
    mocked_courtesy_sleep.assert_not_called()  # 1st request, no sleep
    client.fetch(f'{server}/hello')
    expected_sleep = 5 if courtesy_sleep is None else courtesy_sleep.total_seconds()
    if expected_sleep == 0:
        mocked_courtesy_sleep.assert_not_called()
    else:
        mocked_courtesy_sleep.assert_called_once()
        delay, = mocked_courtesy_sleep.call_args[0]
        assert delay == pytest.approx(expected_sleep, 0.1)


def test_method_kwarg_overrides_default(mocked_courtesy_sleep, server):
    client = Client()
    client.fetch(f'{server}/hello')
    mocked_courtesy_sleep.assert_not_called()  # 1st request, no sleep
    client.fetch(f'{server}/hello', courtesy_sleep=timedelta(minutes=1))
    mocked_courtesy_sleep.assert_called_once()
    assert mocked_courtesy_sleep.call_args[0][0] == pytest.approx(60, 0.1)


def test_method_kwarg_zero_disables_courtesy_sleep(mocked_courtesy_sleep, server):
    client = Client()
    client.fetch(f'{server}/hello')
    mocked_courtesy_sleep.assert_not_called()  # 1st request, no sleep
    client.fetch(f'{server}/hello', courtesy_sleep=timedelta(0))
    mocked_courtesy_sleep.assert_not_called()


def test_method_kwarg_zero_none_does_nothing(mocked_courtesy_sleep, server):
    client = Client()
    client.fetch(f'{server}/hello')
    mocked_courtesy_sleep.assert_not_called()  # 1st request, no sleep
    client.fetch(f'{server}/hello', courtesy_sleep=None)
    mocked_courtesy_sleep.assert_called_once()
    assert mocked_courtesy_sleep.call_args[0][0] == pytest.approx(5, 0.1)  # slept default 5 seconds


def test_nonequal_hostnames(mocker, mocked_courtesy_sleep):
    client = Client()
    mocker.patch.object(client.session, 'request', return_value=dummy_response(dummy_prepared_request(client)))
    client.fetch('http://one/')
    mocked_courtesy_sleep.assert_not_called()
    client.fetch('http://two/')
    mocked_courtesy_sleep.assert_not_called()


@pytest.mark.parametrize(
    'url_1, url_2',
    combinations(
        [
            # All of these are the same host, and so we should sleep between fetches of any pair of URLs in here
            'http://test/',
            'http://TEST/',
            'https://test/',
            'https://TEST/',
            'http://test:8080',
        ],
        2,
    )
)
def test_equal_hostnames(mocker, mocked_courtesy_sleep, url_1, url_2):
    client = Client()
    mocker.patch.object(client.session, 'request', return_value=dummy_response(dummy_prepared_request(client)))
    client.fetch(url_1)
    mocked_courtesy_sleep.assert_not_called()  # 1st request, no sleep
    client.fetch(url_2)
    mocked_courtesy_sleep.assert_called_once()



#!/usr/bin/env python3

# standards
from abc import ABC

# 3rd parties
from requests import RequestException


class ScraperError(ABC, Exception):
    pass


# These are the exceptions that @scraper handles by default. We register them as subclasses of `ScraperError`, as a convenience to
# user code, so that user code doesn't need to import requests just for that exception class.
#
ScraperError.register(ValueError)
ScraperError.register(RequestException)
